# code: utf-8
import re
import os
import sys
import json
import string
import logging
import subprocess

import click

from jmb.tools import logger, configure_logging
from jmb.tools.register import client
try:
    from logutils.colorize import ColorizingStreamHandler as StreamHandler
    # FIXME: colora debug ma non warn!!!
except ImportError:
    from logging import StreamHandler
from jmb.tools.utils.functions import execute, execute0


BARMAN_IP = '212.31.253.42'


class Opts(object):
    pass
opts = Opts()

# -l, --debug-level=LEVEL: set logging level
# -v, --verbose: be verbose
# -n, --dry-run: don't really execute


@click.command('barman')
@click.option('-p', '--local-db-port', default='5432')
@click.option('-B', '--barman-server', default='hgdb-backup.thundersystems.it')
@click.option('-s', '--ssh-port', default='22', help='set port for ssh (3222 per backup 32 bit)')
@click.option('-P', '--password', help='set password for user postgresql')
@click.argument('name', nargs=1)
@click.pass_context
def barman(ctx, name, local_db_port, barman_server, ssh_port, password):
    """
    Configura postgres per barman. Usage examples:

      \b
      jmb conf pg barman -s 3222 publieditor
      jmb conf pg barman thunder-dev

    Add "archive_timeout = 120"  to force a wal any 2 minutes

    NOTE: the cluster is looked up from port

    In barman server:

      \b
      * add the db in /etc/barman.d
        -> fix della porta del pg_cluster
      * check: 'sudo -Hu barman ssh postgres@trepalchi.it date'
      * check: 'barman check <server-name>'
      * barman backup <server-name>


    pyrewall restart per accedere a postgresq sulle porte implicate

    """

    global opts, SUBST
    opts.port = local_db_port
    opts.backup_server = barman_server
    opts.ssh_port = ssh_port
    opts.ssh_opts = "-p " + ssh_port if ssh_port else ''
    opts.password = password
    opts.name = name
    # opts.verbose = ctx.parent.parent.parent.verbose
    # if opts.verbose:
    # I want to issue informations always!!!
    configure_logging('jmb', level=ctx.parent.parent.parent.level, console=True)
    opts.dry_run = ctx.parent.parent.parent.dry_run
    opts.cluster = get_cluster_from_port(port=opts.port)
    opts.pg_conf = '/etc/postgresql/{cluster}/postgresql.conf'.format(**vars(opts))
    if opts.password:
        opts.auth = 'md5'
    else:
        opts.auth = 'trust'
    if re.search('[a-z]', opts.backup_server):
        ret_code, out, err = execute0('host ' + opts.backup_server)
        opts.backup_server = out.split()[-1]

    dest_wals_position = 'barman@{backup_server}:/var/lib/barman/{name}/incoming/%f'.format(
        **vars(opts))
    rsync_opts = '-e "ssh {ssh_opts}"'.format(**vars(opts))
    SUBST = {
        'wal_level': 'replica',
        'listen_addresses': "'*'",
        'archive_mode': 'on',
        'archive_command': "'rsync {rsync_opts} -a %p {dest_wals_position}'".format(**locals())
    }
    create_barman_pguser()
    create_streaming_barman_pguser()
    setup_check()
    fix_pg_conf()
    set_password()
    add_key_barman()
    generate_public_key_and_copy()
    configure_firewall()


def execute0_verb(cmd, *args):

    if isinstance(cmd, list):
        logger.debug(" ".join([repr(x) if ' ' in x else x for x in cmd]))
    else:
        logger.debug(cmd)

    if not opts.dry_run:
        return execute0(cmd, *args)


def get_cluster_from_port(port=5432):

    cluster_line = subprocess.check_output('pg_lsclusters|grep {}'.format(port), shell=True)
    ver, cluster_name, port, status, owner, data_dir, logs = cluster_line.split()
    return "{ver}/{cluster_name}".format(**locals())


def fix_pg_conf():

    logger.info("## Configuring postgresql.conf")
    for key, val in SUBST.iteritems():
        cmd = ['sed', '-i', "/{key}/ c\{key} = {val}".format(key=key, val=val), opts.pg_conf]
        execute0_verb(cmd)


def fix_pg_hba(role, db='all'):

    logger.info("## Configuring pg_hba.conf")
    pg_hba = '/etc/postgresql/{cluster}/pg_hba.conf'.format(cluster=opts.cluster)
    with open(pg_hba) as f:
        txt = f.read()
        line = 'host\t{db}\t{role}\t{backup_server}/32\t{auth} '.format(
            role=role, db=db, **vars(opts))
        if '\t{}'.format(role) not in txt and not opts.dry_run:
            logger.debug(line)
            with open(pg_hba, 'a') as f:
                f.write(line + "\n")
        else:
            # a che serve?
            execute0_verb(['sed', '-i', '/## barman/ c{}'.format(line), pg_hba])


def generate_password(pass_len):
    try:
        # Python 3.6, criptograficaly correct
        import secrets
        alphabet = string.ascii_letters + string.digits
        return ''.join(secrets.choice(alphabet) for i in range(pass_len))
    except:
        from random import choice
        chars = string.letters + string.digits
        return ''.join(choice(chars) for _ in range(pass_len))


def set_password(role='postgres', password=None):
    password = password or opts.password

    if not password:
        return
        # print("Imposta la password con opt -P")
        # sys.exit(1)
    SQL = "ALTER ROLE {role} WITH PASSWORD '{password}'".format(role=role, password=password)
    logger.info("## Adding password for %s", role)
    execute0_verb(["sudo", "-iu", "postgres", "psql", "-p", opts.port, "-c", SQL])


def create_barman_pguser():
    cmd = ['sudo', '-Hu', 'postgres', 'psql', '-p', opts.port, '-c' '\du barman']
    ret, out, err = execute0(cmd)
    if 'barman' not in out:
        execute0_verb(["sudo", "-iu", "postgres", "createuser", '-p', opts.port, "-s", "barman"])
    pwd = opts.password or generate_password(8)
    set_password(role='barman', password=pwd)
    logger.info("Add password '{name}:{port}:*:barman:{password}'".format(
        name=opts.name, port=opts.port, password=pwd)
    )
    fix_pg_hba('barman')


def create_streaming_barman_pguser():
    cmd = ['sudo', '-Hu', 'postgres', 'psql', '-p', opts.port, '-c' '\du streaming_barman']
    ret, out, err = execute0(cmd)
    if 'barman' not in out:
        execute0_verb(
            ["sudo",
                     "-iu",
                     "postgres",
                     "createuser",
                     '-p',
                     opts.port,
                     "--replication",
                     "streaming_barman"])
    pwd = opts.password or generate_password(8)
    set_password(role='streaming_barman', password=pwd)
    logger.info("Add password '{name}:{port}:*:streaming_barman:{password}'".format(
        name=opts.name, port=opts.port, password=pwd)
    )
    fix_pg_hba('streaming_barman', db='replication')


def add_key_barman():

    logger.info("## Granting access as postgres to barman")
    execute0_verb('sudo -Hu postgres add-key barman')


def generate_public_key_and_copy():

    logger.info("## Creating id_rsa")
    rsa_file = os.path.expanduser('~postgres/.ssh/id_rsa')
    cmd = "ssh {ssh_opts} -o StrictHostKeyChecking=no -o LogLevel=ERROR {backup_server} date"
    execute(cmd.format(**vars(opts)))
    # ssh-keygen
    if not os.path.exists(rsa_file):
        cmd = ['sudo', '-u', 'postgres', 'ssh-keygen', '-C', "pg@" + opts.name, '-t', 'rsa',
               '-N', '', '-f', rsa_file]
        execute0(cmd)

    # ssh-copy-id
    cmd = ['ssh-copy-id', '-i', rsa_file, 'barman@' + opts.backup_server]

    if opts.ssh_port:
        cmd.insert(1, '-p')
        cmd.insert(2, opts.ssh_port)
    logger.info("## Copying id_rsa to barman@{}".format(opts.backup_server))
    execute0_verb(cmd)
    # test and known_hosts
    cmd = "sudo -u postgres ssh {ssh_opts} -o StrictHostKeyChecking=no barman@{backup_server} date"
    execute(cmd.format(**vars(opts)))


def set_logger(opts, name='barman'):

    level = opts.debug_level and getattr(logging, opts.debug_level.upper()) or logging.INFO
    logger = logging.getLogger(name)
    console_handler = StreamHandler()
    console_handler.setLevel(level)
    logger.addHandler(console_handler)
    logger.setLevel(level)
    return logger


def configure_firewall():

    logger.info('## Open firewall from backup server  -- FIX if not {}'.format(BARMAN_IP))

    cmd = 'iptb -A ext -s {} -p tcp --dport {} -j ACCEPT\n'.format(BARMAN_IP, opts.port)
    with open('/etc/argo/fw.add', 'r') as f:
        txt = f.read()
        if cmd in txt:
            logger.warn('## Firewall already configured')
            return
    with open('/etc/argo/fw.add', 'a') as f:
        f.write('## barman backup postgresql\n')
        f.write('iptb -A ext -s {} -p tcp --dport {} -j ACCEPT\n'.format(BARMAN_IP, opts.port))

    execute0_verb(['pyrewall', '-r', 'iptb -A ext -s {} -p tcp --dport 5432 -j ACCEPT'.format(
        BARMAN_IP)])
    logger.warn('## Se esiste firewall separato apri con')
    logger.warn('## Se esiste firewall separato apri con')
    logger.warn('   dnat {} <IP>  <IP>  tcp 5432'.format(BARMAN_IP))


def setup_check():

    out = subprocess.check_output('dpkg -l openssh-client|grep openssh-client', shell=True)
    ver = out.split()[2].split(':')[1][:3]
    if ver < '6.6' and not opts.ssh_port == '22':
        logger.error("ATTENZIONE: openssh-client da aggiornare")
        logger.error("deb http://http.debian.net/debian wheezy-backports main")
        logger.error("apt-get -t wheezy-backports  install openssh-client")
        sys.exit(1)


def parse_opts(args, opts):

    if opts.dry_run:
        opts.verbose = True

    if opts.verbose:
        logger.setLevel(logging.DEBUG)

    opts.name = args[0]
    opts.port = opts.port or 5432
    opts.backup_server = opts.backup_server or BARMAN_IP
    if re.search('[a-z]', opts.backup_server):
        ret_code, out, err = execute0('host ' + opts.backup_server)
        opts.backup_server = out.split()[-1]
    if opts.password:
        opts.auth = 'md5'
    else:
        opts.auth = 'trust'

    if opts.ssh_port:
        opts.ssh_opts = "-p {}".format(opts.ssh_port)
    else:
        opts.ssh_opts = ""

    opts.cluster = get_cluster_from_port(port=opts.port)
    opts.pg_conf = '/etc/postgresql/{cluster}/postgresql.conf'.format(**vars(opts))
    return opts


@click.command('backups')
@click.option('-f', '--filename', help="A json file with the result of 'barman diagnose'")
@click.pass_context
def backups(ctx, filename):
    """Send backup information returned bu 'barman diagnose'
    """
    data = json.loads(get_json_data(filename))

    stats_client = client.StatsClient()
    stats_client.backups.bulk_from_barman_diagnose(data['servers'])


def get_json_data(filename):
    if filename:
        data = open(filename).read()
    else:
        cmd = "barman diagnose"
        ret_code, data, err = execute0(cmd)
    return data


@click.command('status')
@click.option('-f', '--filename', help="A json file with the result of 'barman diagnose'")
@click.pass_context
def status(ctx, filename):
    """Send status information returned bu 'barman status'
    """
    data = (get_status_data(filename))

    stats_client = client.StatsClient()
    stats_client.backups.bulk_from_barman_status(data)


def get_status_data(filename):
    if filename:
        data = open(filename).read()
    else:
        cmd = "barman status all"
        ret_code, data, err = execute0(cmd)

    output = {}
    for x in re.split('\nServer ', data, re.MULTILINE | re.DOTALL):
        output.update(get_dict_from_txt(x))
    return output


"""
cermanager:
    Description: Cermanager test machine
    Active: True
    Disabled: False
    PostgreSQL version: 10.2
    Cluster state: in production
    pgespresso extension: Not available
    Current data size: 122.9 MiB
    PostgreSQL Data directory: /var/lib/postgresql/10/main
    Current WAL segment: 000000010000000000000031
    PostgreSQL 'archive_command' setting: rsync -e "ssh -p 22" -a %p barman@212.31.253.42:/var/lib/barman/cermanager/incoming/%f
    Last archived WAL: 000000010000000000000030, at Thu Mar 22 01:10:48 2018
    Failures of WAL archiver: 0
    Server WAL archiving rate: 0.01/hour
    Retention policies: enforced (mode: auto, retention: RECOVERY WINDOW OF 4 WEEKS, WAL retention: MAIN)
    No. of available backups: 8
    First available backup: 20180315T021003
    Last available backup: 20180322T021007
    Minimum redundancy requirements: satisfied (8/0)
"""


def get_dict_from_txt(info):

    data = {}
    lines = info.splitlines()

    for line in lines:
        if 'Skipping inactive' in line:
            return {}
        key, val = re.split(':', line, 1)

        if 'Server WAL archiving rate' in key:
            data['wal_rate'] = val
        elif 'No. of available backups' in key:
            data['n_backups'] = val
        elif 'Current data size' in key:
            data['size'] = val
        elif 'Last available backup' in key:
            data['last_backup'] = val
        elif 'Description' in key:
            data['description'] = val

    return {lines[0].split(':')[0]: data}
