# coding: utf-8

import re
import os
import sys
import glob
from pprint import pprint
from time import sleep
from pprint import pprint

from six.moves import configparser
from .client import StatsClient
from jmb.tools.utils import read
from jmb.tools.conf import configure
from jmb.tools.conf import nginx
from jmb import tools
from jmb.tools import logger
import click


@click.command('project')
@click.option('-b', '--batch', is_flag=True, default=False, help="Don't use progressbar")
@click.option('--host', is_flag=True, default=None, help="Push just info on host and keys")
@click.option('--apps/--no-apps', is_flag=True, default=True, help="if True apps are updated")
@click.option('-d', '--dev', 'settings', flag_value='dev', default=True, help="This is a dev system")
@click.option('-n', '--dry-run', is_flag=True, default=False, help="Just show data, dont' update")
@click.option('-s', '--staging', 'settings', flag_value='staging', default=True, help="This is a staging system")
@click.option('-p', '--production', 'settings', flag_value='production', help="This is a production system")
@click.option('-t', '--test', is_flag=True, help="The project is not yet in production")
@click.option('-P', '--pattern', help="A pattern to filter which packages to push")
@click.option('-m', '--max-depth', type=int, help="the maxdepth level in which a django project is searched (default=2)", default=2)
@click.argument('repositories', nargs=-1)
@click.pass_context
def push_info(ctx, host, apps, settings, test, pattern, repositories,
              max_depth, dry_run, batch):
    """Push info on remote server for each repo in argument (a directory).
    If ``setup.py`` is not found, it's searchd for ``max-depth`` levels deep.
    """
    try:
        user, password = ctx.obj.get('register', 'user'), ctx.obj.get('register', 'password')
    except configparser.NoSectionError:
        configure.create_customization(ctx.obj)
        return

    sc = StatsClient(ctx.obj.get('register', 'api_url'), user, password)

    if host:
        data = read.ServerHost().data
        sc.servers.add_or_update(data)

    cwd = os.getcwd()
    logger.debug('cd %s' % cwd)

    repositories = [nginx.get_all_repositories(repo) for repo in repositories or "."]

    for reps in repositories:
        for repo in reps:
            logger.info("New repo: %s" % repo)
            try:
                os.chdir(repo)
                push_project_info(ctx, pattern, sc, test, apps, dry_run=dry_run, batch=batch)
                os.chdir(cwd)
            except Exception as e:
                logger.warning(u"Critical problem with setup of repo %s" % repo)
                sys.stdout.write(u"Error %s while parsing repo '%s'\n" % (e, repo))


@click.command('host')
@click.option('-n', '--dry-run', is_flag=True, default=False, help="Just show data, dont' update")
@click.pass_context
def push_host(ctx, dry_run):
    """Push info on remote server for this host
    """
    try:
        user, password = ctx.obj.get('register', 'user'), ctx.obj.get('register', 'password')
    except configparser.NoSectionError:
        configure.create_customization(ctx.obj)
        return
    sc = StatsClient(ctx.obj.get('register', 'api_url'), user, password)

    data = read.ServerHost().data
    if ctx.parent.parent.verbose:
        pprint(data)
    if not dry_run:
        sc.servers.add_or_update(data)


def push_project_info(ctx, pattern, sc, test, apps=True, directory=".", dry_run=False, batch=False):
    """Push all info on the project in directory"""

    if not hasattr(ctx.parent.parent, 'verbose'):
        ctx = ctx.parent
    if ctx.parent.parent.verbose > 1:
        click.secho(os.path.basename(os.path.absname(directory)))
    try:
        data = read.DjangoProject().get_project_info(test)
    except read.CollectInfoException as e:
        click.secho("get_project_info raised CollectInfoError: ", nl=False)
        click.secho("{}".format(e.message), fg='red', reverse=True)
        return

    if ctx.parent.parent.verbose:
        click.secho("Update project {}".format(data['name']), fg='magenta')
    if dry_run:
        pprint(data)
    else:
        sc.projects.add_or_update(data)

    if not apps:
        return
    apps = read.DjangoProject().get_packages()
    project_name = data['name']

    def show_me(x):
        try:
            return x[0]
        except:
            return ''

    def loop_over_items(items):

        for app_item in items:
            logger.debug("App_item: %r" % (app_item,))
            app, app_data = app_item
            if pattern:
                if not re.search(pattern, app):
                    continue

            app_data.update({
                'project': project_name,
                'app': app,
            })

            if dry_run:
                print(app_data)
            else:
                sc.appproject.add_or_update_by_slugs(project_name, app, app_data)

    if batch:
        loop_over_items(apps.items())
    else:
        with click.progressbar(apps.items(), item_show_func=show_me) as bar:
            loop_over_items(bar)


@click.command('show')
@click.option('--host/--no-host', is_flag=True, default=None, help="Push just info on host and keys")
@click.option('--apps/--no-apps', is_flag=True, default=True, help="if True apps are updated")
@click.option('--project/--no-project', is_flag=None, default=None,
              help="Push just info on projects and apps")
@click.option('-d', '--dev', 'settings', flag_value='dev', default=True, help="This is a dev system")
@click.option('-s', '--staging', 'settings', flag_value='staging', default=True, help="This is a staging system")
@click.option('-p', '--production', 'settings', flag_value='production', help="This is a production system")
@click.option('-t', '--test', is_flag=True, help="The project is not yet in production")
@click.option('-P', '--pattern', help="A pattern to filter which packages to push")
@click.option('-m', '--max-depth', type=int, help="the maxdepth level in which a django project is searched (default=2)", default=2)
@click.argument('repositories', nargs=-1)
@click.pass_context
def show(ctx, host, apps, project, settings, test, pattern, repositories,
         max_depth):
    """Show all info that will be pushed. Args are directories, project is searched
       ``maxdepth`` level deep
    """
    ctx.forward(push_info, dry_run=True, batch=True)


@click.command('test')
@click.pass_context
def test(ctx):
    """test if the client can connect to the server"""

    user, password = ctx.obj.get('register', 'user'), ctx.obj.get('register', 'password')
    sc = StatsClient(ctx.obj.get('register', 'api_url'), user, password)
    try:
        sc.test()
        click.secho("Ok!", fg="green")
    except Exception as e:
        error = sc.resp.json()['detail']
        click.secho("Not connected! {}".format(error), fg="red")


def parse_flags(a, b):
    if a is None and b is None:
        return True, True
    if a is None and b is True:
        return False, True
    if a is None and b is False:
        return True, False
    if a is True and b is None:
        return True, False
    if a is False and b is None:
        return False, True


@click.command('site')
@click.option('-n', '--dry-run', is_flag=True, default=False, help="Just show data, dont' update")
@click.option('--enabled', '-e', is_flag=True, help="look for certificates in nginx enabled ",
              default=False)
@click.argument('conffiles', nargs=-1)
@click.pass_context
def push_site_info(ctx, conffiles, dry_run, enabled):
    """Push info on remote server for each repo in argument (a directory).
    If ``setup.py`` is not found, it's searched for ``max-depth`` levels deep.
    """
    from jmb.tools.conf import nginx

    if conffiles and enabled:
        msg = u"Opt enabled implies all certificates from sites-enabled!!! no args permitted"
        raise click.BadParameter(msg)

    if enabled:
        conffiles = get_enabled_files()

    user, password = ctx.obj.get('register', 'user'), ctx.obj.get('register', 'password')
    sc = StatsClient(ctx.obj.get('register', 'api_url'), user, password)

    for filename in conffiles:

        if ctx.parent.parent.verbose:
            click.secho("parsing file {}".format(filename), fg='magenta')

        if not os.path.isfile(filename):
            continue

        conf = nginx.ParseConf(filename)
        for srv in conf.servers:
            data = {
                #'root': srv.root,
                'name': srv.names and srv.names[0],
                #'ports': " ".join(srv.ports['http'] + srv.ports['https']),
                'server': read.ServerHost().get_hostname()
            }
            if srv.ssl_certificate:
                data['ssl_certificate'] = srv.ssl_certificate.certificate
                data['ssl_provider'] = srv.ssl_certificate.provider
                data['ssl_expire_date'] = srv.ssl_certificate.expiry.strftime("%Y-%m-%d %H:%M")

                # data['ssl_remaining'] = srv.ssl_certificate.remaining
                data['ssl_alternative_names'] = srv.ssl_certificate.alternative_names.replace(
                    ', ', '\n')
                data['ssl_certificate_name'] = data['name']
                m = re.search(
                    'live/(?P<name>[-\w\._]+)/fullchain.pem',
                    srv.ssl_certificate.certificate)
                if m:
                    data['ssl_certificate_name'] = m.group('name')

            if dry_run or ctx.parent.parent.verbose:
                pprint(data)
            if not dry_run:
                try:
                    sc.domains.create_or_update(data)
                except:
                    print(sc.domains.response.json())


def get_enabled_files():
    """Return all files under NGINX_ENABLED_DIR (/etc/nginx/sites-enabled)
    possibly with recursion
    """
    for root, dirs, files in os.walk(tools.NGINX_ENABLED_DIR, topdown=True):
        for name in files:
            yield name
