 # coding: utf-8
"""
Configuration reporter
======================

Sample output::

  {
    u'buildout': u'2',
    u'celery': False,
    u'debug': u'True',
    u'dj_version': '1.11.3',
    u'name': 'django',
    u'nginx': '/etc/nginx/sites-enabled/whistle.conf',
    u'py_version': u'3',
    u'repo': 'ssh://hg@bitbucket.org/thux/office2017.whistle.it',
    u'sentry': False,
    u'server': 'cli-prod-uk',
    u'settings': u'dev',
    u'test': False
  }



"""
from __future__ import unicode_literals

import io
import os
import re
import six
import ssl
import socket
import subprocess
import datetime
import click
import subprocess as sp
from collections import OrderedDict
from urllib3.util import parse_url

from six.moves import configparser
from .functions import execute, execute0
from OpenSSL import crypto
import hglib


class CollectInfoException(Exception):
    pass

# Configuration


def get_config_filename():
    return os.path.join(click.get_app_dir('jmb'), 'jmb.ini')


def read_config_file(cfg=None):
    if not cfg:
        cfg = get_config_filename()
    parser = six.moves.configparser.SafeConfigParser()
    parser.read([cfg])
    return parser

# Server


class ServerHost(object):

    def __init__(self):
        data = self.get_os()
        data['ip'] = self.get_ip()
        data['name'] = self.get_hostname()
        data['keys'] = self.get_keys()
        data['mailname'] = self.get_mailname()
        data['root_alias'] = self.get_root_alias()
        data['argodef'] = self.get_dpkg_version('argodef')
        data['jmb_tools'] = self.get_dpkg_version('jmb_tools') or self.get_jmb_tools_release()
        self.data = data

    def get_os(self):
        "Return a dict with os, rel, name"

        ret_code, out, err = execute('lsb_release -a')
        info = {}
        for line in out.splitlines():
            line = line.decode('utf-8')
            if line.startswith('Distributor ID'):
                info['distributor'] = line.split('\t')[1]
            elif line.startswith('Release'):
                info['release'] = line.split('\t')[1]
            elif line.startswith('Codename'):
                info['codename'] = line.split('\t')[1]
        return info

    def get_ip(self):
        s = socket.socket()
        s.connect(('google.com', 80))
        return s.getsockname()[0]

    def get_hostname(self):
        return socket.gethostname()

    def get_mailname(self):
        if os.path.exists('/etc/mailname'):
            return open('/etc/mailname').read().strip()

    def get_jmb_tools_release(self):
        ret_code, output, err = execute('jmb --version')
        if ret_code == 0:
            return output.split()[-1]

    def get_dpkg_version(self, pkg):
        ret_code, output, err = execute('dpkg -l {pkg}'.format(pkg=pkg))
        for line in output.splitlines():
            if pkg in line:
                return line.split()[2]

    def get_root_alias(self):
        for line in open('/etc/aliases'):
            if not ':' in line:
                continue
            alias, value = line.split(':')
            if alias == 'root':
                m = re.search('(?P<value>.*)(#.*)?', value)
                if m:
                    return m.group('value').strip()

    def get_keys(self):
        ret_code, out, err = execute('add-key')
        keys = out.decode('utf-8')
        return [line.split()[-1] for line in keys.splitlines()]

    def get_server_info(self):
        return self.data

# Project


class DjangoProject(object):

    def __init__(self, directory="."):
        self.directory = os.path.abspath(directory)

    def get_project(self):
        info = {}
        if not os.path.exists(os.path.join(self.directory, 'setup.py')):
            return info
        # info['packages'] = get_packages(self.directory)
        cp = six.moves.configparser.ConfigParser()
        cp.read(os.path.join(self.directory, '.hg/hgrc'))
        try:
            repo = cp.get('paths', 'default')
            info['name'] = os.path.basename(repo)
        except:
            repo = None
            info['name'] = os.path.basename(self.directory)
        # ret_code, out, err = execute('jmb-test-setup')

        return info

    def get_pyversion(self):
        ret_code, out, err = execute0('head -n 1 bin/buildout', cwd=self.directory)
        return '3' if 'python3' in out.decode('utf-8') else '2'

    def get_djversion(self):
        ret_code, out, err = execute0('jmb-test-setup -VE Django', cwd=self.directory)
        return out.strip()

    def get_buildout_version(self):
        "return a tuple major release and complete rel"
        try:
            ret_code, out, err = execute0("grep zc.buildout.*egg %s/bin/buildout" % self.directory)
        except:
            ret_code, out, err = execute0("grep zc.buildout.*egg %s/bin/django" % self.directory)
        rel = out.decode('utf-8').split('-')[1]
        return rel[0], rel

    def get_packages(self, pattern='.*'):
        cwd = os.getcwd()
        os.chdir(self.directory)

        ret_code, out, err = execute("jmb-test-setup")
        packages = OrderedDict()
        for line in out.decode('utf-8').splitlines():
            if not re.search(pattern, line):
                continue
            m = re.search('/(?P<name>[\w\._]+)-(?P<version>[\d\.\w]+)-py.*.egg', line)
            if m:
                info = {'egg': True}
                info.update(m.groupdict())
                packages[m.groupdict()['name']] = info
            else:
                pkg = line.split('/')[-1]
                if not pkg:
                    click.secho("Discarding line: {}".format(line), fg='yellow', reverse=True)
                    continue
                packages[pkg] = self.get_revision_info(line)

        os.chdir(cwd)
        return packages

    def get_revision_info(self, repo):
        info = {}
        if not os.path.exists(os.path.join(repo, '.hg')):
            return {'rev': None, 'egg': False}
        cli = hglib.open(repo)
        rev = cli.parents()[0]
        info['rev'] = rev.rev
        info['branch'] = rev.branch
        info['node'] = rev.node[:12]
        info['egg'] = False

        return info

    def get_debug(self):
        ret_code, out, err = execute('bin/django diffsettings')
        for line in out.decode('utf-8').splitlines():
            if line.startswith('DEBUG ='):
                return line.split(' = ')[1]
        return False

    def get_sentry(self):
        ret_code, out, err = execute('bin/django diffsettings')
        for line in out.decode('utf-8').splitlines():
            if line.startswith('RAVEN_CONFIG ='):
                return line.split(' = ')[1] and True or False
        return False

    def get_settings_link(self):

        settings_file = os.path.join(self.directory, 'web/settings/settings.py')

        if not os.path.exists(settings_file) or not os.path.islink(settings_file):
            return
        return os.readlink(settings_file).split('.')[0]

    def get_celery(self):
        nick_file = os.path.join(self.directory, 'bin/nick')
        if not os.path.exists(nick_file):
            return False
        with io.open(nick_file) as f:
            nick = f.readline().split()[0]
        ini_file = '/etc/uwsgi-emperor/vassals-enabled/{nick}.ini'.format(nick=nick)
        ini_file2 = '/etc/uwsgi/{nick}.ini'.format(nick=nick)
        if os.path.exists(ini_file):
            pass
        elif os.path.exists(ini_file2):
            ini_file = ini_file2
        else:
            return None
        cp = six.moves.configparser.RawConfigParser()
        cp.read(ini_file)
        try:
            return cp.get('uwsgi', 'attach-daemon2') and True or False
        except configparser.NoOptionError:
            return False

    def get_nginx_conf(self):
        """Return the nginx file searching first in .jmb.conf and then in register
        """
        from jmb.tools import conf
        jmb_conf = os.path.join(self.directory, '.jmb.conf')

        if os.path.exists(jmb_conf):
            with open(jmb_conf) as f:
                for line in f:
                    m = re.search('nginx_conf\s=\s(?P<filename>[-\w\./]+)', line)
                    if m:
                        return m.group('filename')
        return conf.get_nginx_conf(self.directory)

    def get_project_info(self, test=False):

        repo_dir = os.path.basename(self.directory)
        if not os.path.exists(os.path.join(self.directory, 'buildout.cfg')):
            raise CollectInfoException("Repo {} does not seem a buildout project".format(repo_dir))

        if not os.path.exists(os.path.join(self.directory, 'bin/buildout')):
            raise CollectInfoException(
                "Repo {} does not seem to be initialized (jmb-go?)".format(repo_dir))

        data = self.get_project()
        data['test'] = test
        data['settings'] = self.get_settings_link()
        data['server'] = ServerHost().get_hostname()
        data['py_version'] = self.get_pyversion()
        data['dj_version'] = self.get_djversion()
        data['buildout'] = self.get_buildout_version()[0]
        data['debug'] = self.get_debug()
        data['celery'] = self.get_celery()
        data['sentry'] = self.get_sentry()
        data['nginx'] = self.get_nginx_conf()
        # data['ports'] = self.get_ports()
        # data['expiry'] = self.get_expiry()
        return data


# Functions

def read_certificate(cert_file):
    cert = crypto.load_certificate(crypto.FILETYPE_PEM, open(cert_file).read())
    expiry = datetime.datetime.strptime(cert.get_notAfter(), "%Y%m%d%H%M%SZ")
    remaining = expiry - datetime.datetime.utcnow()
    domain = cert.get_subject().get_components()[0][1]
    alternative_names = get_alternative_names(cert_file)
    return domain, expiry, remaining, alternative_names


def get_alternative_names(cert_file):
    cmd = 'openssl x509 -in {} -noout -text'.format(cert_file)
    p = sp.Popen(cmd.split(), stdin=sp.PIPE, stdout=sp.PIPE)
    txt = p.stdout.read()
    for line in txt.splitlines():
        if 'DNS' in line:
            return line.strip().replace('DNS:', '')


def ssl_expiry_datetime(hostname):
    ssl_date_fmt = r'%b %d %H:%M:%S %Y %Z'

    context = ssl.create_default_context()
    conn = context.wrap_socket(
        socket.socket(socket.AF_INET),
        server_hostname=hostname,
    )
    # 3 second timeout because Lambda has runtime limitations
    conn.settimeout(3.0)

    host = parse_url(hostname)
    try:
        conn.connect((host.host, host.port or 443))
    except ssl.SSLError:
        return "Expired"
    ssl_info = conn.getpeercert()
    # parse the string from the certificate into a Python datetime object
    return datetime.datetime.strptime(ssl_info['notAfter'], ssl_date_fmt)


def ssl_valid_time_remaining(hostname):
    """Get the number of days left in a cert's lifetime."""
    expires = ssl_expiry_datetime(hostname)
    if expires == 'Expired':
        return 'Expired'
    return expires - datetime.datetime.utcnow()
